#include "Encryption.h"
#include "Conf.h"

#define ENC_SAFE_EXIT(x) 	if(hDstFile != INVALID_HANDLE_VALUE) CloseHandle(hDstFile); \
							if(hSrcFile != INVALID_HANDLE_VALUE) CloseHandle(hSrcFile); \
							DeleteFile(pOutFile); \
							if(pRead) delete[] pRead; \
							if(pWrite) delete[] pWrite; \
							return x;

#define DEC_SAFE_EXIT(x) 	if(hDstFile != INVALID_HANDLE_VALUE) CloseHandle(hDstFile); \
							if(hSrcFile != INVALID_HANDLE_VALUE) CloseHandle(hSrcFile); \
							DeleteFile(*pOutFile); \
							if(pRead) delete[] pRead; \
							if(pWrite) delete[] pWrite; \
							if(*pOutFile) delete[] (*pOutFile); \
							return x;

#define MEMDEC_SAFE_EXIT(x) if(hSrcFile != INVALID_HANDLE_VALUE) CloseHandle(hSrcFile); \
							if(pRead) delete[] pRead; \
							return x;

#define ALPHABET_LEN 64

Encryption::Encryption(BYTE *pKey, UINT KeyLen){
	// Possiamo accettare solo queste tre lunghezze per la chiave
	if(KeyLen != 128 && KeyLen != 192 && KeyLen != 256)
		return;

	// La lunghezza della chiave deve essere in bit (128/192/256)
	aes_set_key(&Aes_ctx, pKey, KeyLen);

	// L'IV sempre a 0 per la scrittura in streaming dei log
	ZeroMemory(&IV, 16);
}

/*BOOL Encryption::EncryptFile(WCHAR* pInFile, WCHAR* pOutFile, UINT Type) {
	BYTE *pRead = NULL, *pWrite= NULL, t_IV[16];
	DWORD dwRead = 0, dwWritten = 0, dwFileSize = 0;
	UINT uChunks = 0, i;
	HANDLE hSrcFile = INVALID_HANDLE_VALUE;
	HANDLE hDstFile = INVALID_HANDLE_VALUE;
	SYSTEMTIME st;
	FILETIME ft;
	LogStruct LogDescription;

	memcpy(t_IV, IV, 16);

	// Apriamo il file da cifrare
	hSrcFile = CreateFile(pInFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hSrcFile == INVALID_HANDLE_VALUE)
		return FALSE;

	// Prendiamo la dimensione originale del file
	dwFileSize = GetFileSize(hSrcFile, NULL);

	if(dwFileSize == INVALID_FILE_SIZE || dwFileSize == 0){
		CloseHandle(hSrcFile);
		return FALSE;
	}

	// Creiamo il file di destinazione
	hDstFile = CreateFile(pOutFile, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hDstFile == INVALID_HANDLE_VALUE){
		CloseHandle(hSrcFile);
		return FALSE;
	}

	GetSystemTime(&st);
	SystemTimeToFileTime(&st, &ft);

	// Riempiamo i membri della struttura
	LogDescription.cbSize = sizeof(LogStruct);
	LogDescription.uVersion = LOG_VERSION_01;
	LogDescription.uHTimestamp = ft.dwHighDateTime;
	LogDescription.uLTimestamp = ft.dwLowDateTime;
	LogDescription.FileNameLength = WideLen(pOutFile);

	// Scrivi l'header nel file
	if(WriteFile(hDstFile, (BYTE *)&LogDescription, sizeof(LogStruct), &dwWritten, NULL) == FALSE){
		ENC_SAFE_EXIT(FALSE)
	}

	// Scriviamo il nome del file
	if(WriteFile(hDstFile, (BYTE *)pOutFile, LogDescription.FileNameLength, &dwWritten, NULL) == FALSE){
		ENC_SAFE_EXIT(FALSE)
	}

	if(dwFileSize > GetNextMultiple(MAX_ALLOCABLE_MEMORY)){
		uChunks = dwFileSize / GetNextMultiple(MAX_ALLOCABLE_MEMORY);
		uChunks++;
	}

	// Scriviamo in vari pezzi se il file e' grande piu' di MAX_ALLOCABLE_MEMORY
	if(uChunks){
		pRead = new(std::nothrow) BYTE[GetNextMultiple(MAX_ALLOCABLE_MEMORY)];
		pWrite = new(std::nothrow) BYTE[GetNextMultiple(MAX_ALLOCABLE_MEMORY)];

		if(pRead == NULL || pWrite == NULL){
			ENC_SAFE_EXIT(FALSE)
		}

		for(i = 0; i < uChunks - 1; i++){
			if(ReadFile(hSrcFile, pRead, GetNextMultiple(MAX_ALLOCABLE_MEMORY), &dwRead, NULL) == FALSE){
				ENC_SAFE_EXIT(FALSE)
			}

			aes_cbc_encrypt(&Aes_ctx, (BYTE *)&t_IV, pRead, pWrite, dwRead);

			if(WriteFile(hDstFile, pWrite, GetNextMultiple(dwRead), &dwWritten, NULL) == FALSE){
				ENC_SAFE_EXIT(FALSE)
			}
		}

		dwFileSize -= (uChunks - 1) * GetNextMultiple(MAX_ALLOCABLE_MEMORY);
	}

	// Se sono gia allocati non serve riallocarli perche' sono gia' grandi MAX_ALLOCABLE_MEMORY
	if(pRead == NULL)
		pRead = new(std::nothrow) BYTE[GetNextMultiple(dwFileSize)];

	if(pWrite == NULL)
		pWrite = new(std::nothrow) BYTE[GetNextMultiple(dwFileSize)];

	if(pRead == NULL || pWrite == NULL){
		ENC_SAFE_EXIT(FALSE)
	}

	if(ReadFile(hSrcFile, pRead, dwFileSize, &dwRead, NULL) == FALSE){
		ENC_SAFE_EXIT(FALSE)
	}

	aes_cbc_encrypt(&Aes_ctx, (BYTE *)&t_IV, pRead, pWrite, dwRead);

	if(WriteFile(hDstFile, pWrite, GetNextMultiple(dwRead), &dwWritten, NULL) == FALSE){
		ENC_SAFE_EXIT(FALSE)
	}

	if(pRead)
		delete[] pRead;

	if(pWrite)
		delete[] pWrite;

	CloseHandle(hSrcFile);
	CloseHandle(hDstFile);
	return TRUE;
}*/

BYTE* Encryption::EncryptData(BYTE *pIn, UINT *Len) {
	BYTE *pOut = NULL;
	BYTE t_IV[16];
	UINT paddedLen;

	paddedLen = GetNextMultiple(*Len);
	pOut = new(std::nothrow) BYTE[paddedLen];

	if(pOut == NULL)
		return NULL;

	memcpy(t_IV, IV, 16);
	*Len = paddedLen;

	aes_cbc_encrypt(&Aes_ctx, (BYTE *)&t_IV, pIn, pOut, *Len);

	return pOut;
}

BYTE* Encryption::DecryptConf(WCHAR* pInFile, UINT *uLen) {
	DWORD dwRead = 0, dwWritten = 0, dwFileSize = 0, dwDataLen, dwUnpadded, crc;
	BYTE *pRead = NULL, *pWrite = NULL, t_IV[16];
	HANDLE hConfFile = INVALID_HANDLE_VALUE;
	WCHAR *completePath = NULL;

	/**
	 * Struttura del file di configurazione
	 *
	 * |DWORD|DWORD|DWORD|DATA.....................|CRC|
	 * |---Skip----|-Len-|
	 *
	 * Le prime due DWORD vanno skippate.
	 * La terza DWORD contiene la lunghezza del blocco di dati
	 * CRC e' il CRC dei dati in chiaro, inclusa la DWORD Len
	 */

	memcpy(t_IV, IV, 16);

	completePath = GetCurrentPath(pInFile);

	if(completePath == NULL)
		return FALSE;

	// Apriamo il file di configurazione
	hConfFile = CreateFile(completePath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	delete[] completePath;

	if(hConfFile == INVALID_HANDLE_VALUE)
		return NULL;
	
	// Prendiamo la dimensione del file cifrato
	dwFileSize = GetFileSize(hConfFile, NULL);

	if(dwFileSize == INVALID_FILE_SIZE || dwFileSize == 0  || dwFileSize < sizeof(UINT) * 3){
		CloseHandle(hConfFile);
		return NULL;
	}

	// Allochiamo memoria per i dati
	dwDataLen = dwFileSize - (sizeof(UINT) * 2); // dwDataLen e' la dimensione di TUTTA la parte cifrata
	pRead = new(std::nothrow) BYTE[GetNextMultiple(dwDataLen)];

	if(pRead == NULL){
		CloseHandle(hConfFile);
		return NULL;
	}

	// Skippiamo le prime due DWORD
	SetFilePointer(hConfFile, sizeof(UINT) * 2, 0, FILE_BEGIN);

	// Leggiamo il blocco dati prima di decifrarlo
	if(ReadFile(hConfFile, pRead, GetNextMultiple(dwDataLen), &dwRead, NULL) == FALSE){
		delete[] pRead;
		CloseHandle(hConfFile);
		return NULL;
	}

	if(dwRead != GetNextMultiple(dwDataLen)){
		delete[] pRead;
		CloseHandle(hConfFile);
		return NULL;
	}

	// E decifriamolo
	aes_cbc_decrypt(&Aes_ctx, (BYTE *)&t_IV, pRead, pRead, GetNextMultiple(dwDataLen));

	// Leggiamo la DWORD che ci dice quanto e' lungo il blocco dati non paddato
	dwUnpadded = *(UINT *)pRead;

	if(GetNextMultiple(dwUnpadded) != GetNextMultiple(dwDataLen)){
		delete[] pRead;
		CloseHandle(hConfFile);
		return NULL;
	}

	// Prendiamo il CRC (con una memcopy perche' l'accesso DEVE essere allineato a 4)
	//crc = *((DWORD *)(pRead + dwUnpadded - 4));
	memcpy(&crc, pRead + dwUnpadded - 4, 4);

	// Controlla il CRC
	DWORD conf_hash;
	LONG64 temp_hash = 0;
	CHAR *ptr = (CHAR *)pRead;
	for (UINT i = 0; i < dwUnpadded - sizeof(UINT); i++) {
		temp_hash++;

		if (*ptr)
			temp_hash *= (*ptr);

		conf_hash = (DWORD)(temp_hash >> 32);
		temp_hash &= 0xFFFFFFFF;
		temp_hash ^= conf_hash;
		ptr++;
	}

	conf_hash = (DWORD)temp_hash;

	if(crc != conf_hash){
		delete[] pRead;
		CloseHandle(hConfFile);
		return NULL;
	}

	// Controlla che ci sia il delimitatore di fine configurazione
	// XXX - Se viene messo lo 0 dopo ENDOF_CONF_DELIMITER dobbiamo tenere il - 1 altrimenti
	// dobbiamo levarlo
	BYTE *confEnd = (BYTE *)(pRead + dwUnpadded - sizeof(UINT) - strlen(ENDOF_CONF_DELIMITER) - 1);
	
	if(memcmp(confEnd, ENDOF_CONF_DELIMITER, strlen(ENDOF_CONF_DELIMITER))){
		delete[] pRead;
		CloseHandle(hConfFile);
		return NULL;
	}

	CloseHandle(hConfFile);

	*uLen = dwUnpadded - 4; // Torna la lunghezza del buffer dati senza CRC
	return pRead;
}

// XXX - Non serve perche' non dobbiamo decifrare mai nulla in memoria
BYTE* Encryption::DecryptFileInMemory(WCHAR* pInFile, UINT *Len) {
	return NULL;
	/*LogStruct LogDescription;
	DWORD dwRead = 0, dwFileSize = 0;
	BYTE *pRead = NULL;
	HANDLE hSrcFile = INVALID_HANDLE_VALUE;

	// Apriamo il file da cifrare
	hSrcFile = CreateFile(pInFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hSrcFile == INVALID_HANDLE_VALUE)
		return NULL;

	// Prendiamo la dimensione del file cifrato
	dwFileSize = GetFileSize(hSrcFile, NULL);

	if(dwFileSize == INVALID_FILE_SIZE || dwFileSize == 0 || dwFileSize < sizeof(LogStruct)){
		CloseHandle(hSrcFile);
		return NULL;
	}

	// Riempiamo la struttura LogStruct
	ReadFile(hSrcFile, &LogDescription.cbSize, 4, &dwRead, NULL);
	ReadFile(hSrcFile, &LogDescription.IV, 16, &dwRead, NULL);
	ReadFile(hSrcFile, &LogDescription.uHTimestamp, 4, &dwRead, NULL);
	ReadFile(hSrcFile, &LogDescription.uLTimestamp, 4, &dwRead, NULL);
	ReadFile(hSrcFile, &LogDescription.UnpaddedFileSize, 4, &dwRead, NULL);
	ReadFile(hSrcFile, &LogDescription.PaddedFileSize, 4, &dwRead, NULL);
	ReadFile(hSrcFile, &LogDescription.FileNameLength, 4, &dwRead, NULL);

	// XXX - Limite sul size del decrypt imposto artificialmente da noi!!!
	// Usciamo anche se il file e' troppo grande, per il momento non implementiamo il decrypt a chunk perche'
	// sul client non dobbiamo decifrare mai nulla ad eccezione della configurazione
	if(LogDescription.cbSize != sizeof(LogDescription) || LogDescription.PaddedFileSize > MAX_ALLOCABLE_MEMORY){
		MEMDEC_SAFE_EXIT(NULL)
	}

	// Skippiamo il nome del file
	SetFilePointer(hSrcFile, LogDescription.FileNameLength, NULL, FILE_CURRENT);

	pRead = new(std::nothrow) BYTE[GetNextMultiple(LogDescription.PaddedFileSize)];

	if(pRead == NULL){
		MEMDEC_SAFE_EXIT(NULL)
	}

	// Prepariamoci per il decrypt
	memcpy(IV, LogDescription.IV, 16);
	ReadFile(hSrcFile, pRead, GetNextMultiple(LogDescription.PaddedFileSize), &dwRead, NULL);
	aes_cbc_decrypt(&Aes_ctx, (BYTE *)&LogDescription.IV, pRead, pRead, GetNextMultiple(LogDescription.PaddedFileSize));
	*Len = LogDescription.UnpaddedFileSize;
	CloseHandle(hSrcFile);

	return pRead;*/
}

WCHAR* Encryption::EncryptName(WCHAR* wName, BYTE seed) {
	return Scramble(wName, seed, TRUE);
}

WCHAR* Encryption::DecryptName(WCHAR* wName, BYTE seed) {
	return Scramble(wName, seed, FALSE);
}

WCHAR* Encryption::Scramble(WCHAR* wName, BYTE seed, BOOL enc) {
	WCHAR *ret_string;
	UINT i, j;

	WCHAR alphabet[ALPHABET_LEN] = {'_','B','q','w','H','a','F','8','T','k','K','D','M',
		'f','O','z','Q','A','S','x','4','V','u','X','d','Z',
		'i','b','U','I','e','y','l','J','W','h','j','0','m',
		'5','o','2','E','r','L','t','6','v','G','R','N','9',
		's','Y','1','n','3','P','p','c','7','g','-','C'};                  

	if (!(ret_string = _wcsdup(wName)))
		return NULL;

	// Evita di lasciare i nomi originali anche se il byte e' 0
	seed = (seed > 0) ? seed %= ALPHABET_LEN : seed;

	if(seed == 0)
		seed = 1;

	for(i = 0; ret_string[i]; i++) {
		for(j = 0; j < ALPHABET_LEN; j++){
			if(ret_string[i] == alphabet[j]) {
				// Se crypt e' TRUE cifra, altrimenti decifra
				if (enc)
					ret_string[i] = alphabet[(j + seed) % ALPHABET_LEN];
				else
					ret_string[i] = alphabet[(j + ALPHABET_LEN - seed) % ALPHABET_LEN];
				break;
			}
		}
	}

	return ret_string;
}

void Encryption::Sha1(BYTE* pInData, UINT len, UINT pSha[5]) {
	SHA1Context shcxt;

	SHA1Reset(&shcxt);
	SHA1Input(&shcxt, pInData, len);
	SHA1Result(&shcxt);

	memcpy(pSha, shcxt.Message_Digest, 20);
}

UINT Encryption::GetNextMultiple(UINT Number) {
#define BLOCK_SIZE 16	// AES Block Size

	UINT remainder = 0;

	if(Number == 0)
		return 0;

	if(Number % BLOCK_SIZE == 0)
		return Number;

	remainder = Number % BLOCK_SIZE;

	return Number + (BLOCK_SIZE - remainder);
}
