#include "Common.h"
#include <string>
#include <exception>
using namespace std;

#ifndef __Status_h__
#define __Status_h__

#define MAX_WAIT_TIME 1000 // Tempo massimo che attendiamo perche' un agente si fermi

class Status
{
	private: list<AgentStruct*> *agents;
	private: list<EventStruct*> *events;
	private: list<MacroActionStruct*> *actions;
	private: list<ConfStruct*> *config;
	private: static Status *Instance;
	private: DeviceStruct *device;

	/**
	 * Semaforo per l'accesso alla lista degli eventi, e' pubblico per avere una scrittura/lettura atomica.
	 */
	public: HANDLE hEventsLock;

	/**
	* Semaforo per l'accesso alla lista delle azioni, e' pubblico per avere una scrittura/lettura atomica.
	*/
	public: HANDLE hActionsLock;

	/**
	 * Semaforo per l'accesso alla lista degli agenti, e' pubblico per avere una scrittura/lettura atomica.
	 */
	public: HANDLE hAgentsLock;

	/**
	* Semaforo per l'accesso alla struttura di configurazione.
	*/
	public: HANDLE hConfigLock;

	public: list<AgentStruct*>* GetAgentsList();

	public: list<EventStruct*>* GetEventsList();

	public: list<MacroActionStruct*>* GetActionsList();

	/**
	* Torna un puntatore alla struttura che definisce i parametri di configurazione.
	*/
	public: list<ConfStruct*>* GetConfigStruct();

	/**
	* Costruttore di default
	*/
	public: Status();
			~Status();

	/**
	* Status e' una classe singleton
	*/
	public: static Status* self();

	/**
	 * Torna un puntatore alla struttura che definisce i parametri specifici del device.
	 */
	public: DeviceStruct* GetDeviceStruct();

	/**
	 * Aggiunge un agente alla lista o ne aggiorna lo stato se l'agente gia' esiste. Torna TRUE se l'agente e' stato
	 * aggiunto/updatato, FALSE altrimenti.
	 */
	public: BOOL AddAgent(AgentStruct* pAgent);

	/**
	 * Aggiunge un evento alla lista o ne aggiorna lo stato se l'evento gia' esiste. Torna TRUE se l'evento e' stato
	 * aggiunto/updatato, FALSE altrimenti.
	 */
	public: BOOL AddEvent(EventStruct* pEvent);

	/**
	 * Aggiunge una lista di azioni alla lista. Torna TRUE se l'azione e' stata aggiunta, FALSE altrimenti.
	 */
	public: BOOL AddAction(MacroActionStruct* pAction);

	/**
	 * Legge i parametri di configurazione dal file nella sezione "configurazione" e popola la ConfStruct. Torna TRUE
	 * se i dati sono stati aggiunti con successo, FALSE altrimenti.
	 */
	public: BOOL AddConf(ConfStruct* pConf);

	/**
	* Inizializza la struttura DeviceStruct con tutte le informazioni specifiche del device. Torna TRUE se la funzione
	* e' andata a buon fine, FALSE altrimenti.
	*/
	public: BOOL SetDeviceStruct(DeviceStruct *dev);

	/**
	 * Svuota la lista degli agenti, degli eventi e delle azioni e tutti i puntatori allocati. 
	 * Torna TRUE se la funzione e' andata a buon fine, FALSE altrimenti. Non va usata prima di aver 
	 * fermato gli agenti!!!!
	 */
	public: BOOL Clear();

	/**
	 * Controlla lo stato dell'agente, Type identifica l'agente da interrogare, puo' tornare
	 * uno di questi due valori: AGENT_STOPPED, AGENT_RUNNING
	 */
	public: UINT QueryAgentStatus(UINT Type);

	/**
	* Controlla lo stato del monitor, Type identifica il monitor da interrogare, puo' tornare
	* uno di questi due valori: EVENT_STOPPED, EVENT_RUNNING
	*/
	public: UINT QueryEventStatus(UINT Type, UINT uActionId);
};

#endif
